<?php

namespace VerifoneEcomAPI\ApiWrapper\Http\ApiClient;

use Exception;
use VerifoneEcomAPI\ApiWrapper\Authentication\Interfaces\AuthenticationInterface;
use VerifoneEcomAPI\ApiWrapper\Http\HttpException;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Regions\AbstractRegion;
use VerifoneEcomAPI\ApiWrapper\Settings;

/**
 * Class Client
 * @package VerifoneEcomAPI\ApiWrapper\Http\ApiClient
 */
class Client {

    private $region;
    private $simpleCurl;
    private $settings;
    private $authentication;
    private $auth;

    /**
     * Client constructor.
     *
     * @param AbstractRegion $region
     * @param Settings $settings
     * @param AuthenticationInterface $authentication
     * @param SimpleCurl $simpleCurl
     */
    public function __construct(
        AbstractRegion $region,
        Settings $settings,
        AuthenticationInterface $authentication,
        SimpleCurl $simpleCurl
    ) {
        $this->region         = $region;
        $this->authentication = $authentication;
        $this->settings       = $settings;
        $this->simpleCurl     = $simpleCurl;
        $this->auth           = $this->authentication->getAuth();
    }

    /**
     * @param array $payload
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     */
    public function postCapture( array $payload, $transactionId ) {


        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postCaptureUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     *
     * @return mixed
     * @throws HttpException
     */
    public function postSubscription( array $payload ) {

        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postSubscriptionUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     */
    public function postPaypalCapture( array $payload, $transactionId ) {


        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postPaypalCaptureUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     */
    public function postVoidAuthorization( array $payload, $transactionId ) {


        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postVoidAuthorizationUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     */
    public function postVoidCapture( $transactionId ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postVoidCaptureUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( [] ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     *
     * @return mixed
     * @throws HttpException
     */
    public function postCustomer( array $payload ) {


        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postCustomerUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     *
     * @return mixed
     * @throws HttpException
     * @throws \Exception
     */
    public function postCheckout( array $payload ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postCheckoutUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     * @throws \Exception
     */
    public function postRefund( array $payload, $transactionId ) {

        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getRefundUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     * @throws \Exception
     */
    public function postPaypalRefund( array $payload, $transactionId ) {

        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getRefundPaypalUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param $checkoutId
     *
     * @return mixed
     * @throws HttpException
     */
    public function getCheckout( $checkoutId ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getCheckoutUrl( $checkoutId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param $transactionId
     *
     * @return mixed
     * @throws HttpException
     */
    public function getTransaction( $transactionId ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getTransactionUrl( $transactionId ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function getKeys() {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getKeyUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function generateKey() {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getKeyUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function generateJwt( string $threedsContractId ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getJwtUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( [
                                       'threeds_contract_id' => $threedsContractId,
                                   ] ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function threeDSLookup( array $payload ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getThreeDSLookupUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function threeDSJwtValidate( array $payload ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getThreeDSJwtValidateUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param array $payload
     *
     * @return mixed
     * @throws HttpException
     * @throws \Exception
     */
    public function postCard( array $payload ) {
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->postCardUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $payload ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @param $client
     */
    private function addSsl( $client ) {
        if ( ! $this->settings->getCurlVerifySsl() ) {
            $client->setOpt( CURLOPT_SSL_VERIFYHOST, false )
                   ->setOpt( CURLOPT_SSL_VERIFYPEER, false );
        }
    }

    /**
     * @return string[]
     * @throws Exception
     */
    private function getHeaders() {

        $headers = [
            'Content-Type: application/json',
            'Authorization: Basic ' . $this->auth,
        ];
        if ($this->simpleCurl->needsVfiApiKey()) {
            $headers[] ='x-vfi-api-idempotencykey: '.  $this->generateUUID4();
        }
        return $headers;
    }

    /**
     * @throws Exception
     */
    private function generateUUID4()
    {
        $data = random_bytes(16);
        assert(strlen($data) == 16);
        $data[6] = chr(ord($data[6]) & 0x0f | 0x40);
        $data[8] = chr(ord($data[8]) & 0x3f | 0x80);
        return vsprintf('%s%s-%s-%s-%s-%s%s%s', str_split(bin2hex($data), 4));
    }

    /**
     * @param $result
     *
     * @return mixed
     * @throws HttpException
     */
    private function decodeResult( $result ) {
        $result = json_decode( $result, true );
        if ( json_last_error() !== JSON_ERROR_NONE ) {
            throw new HttpException( 'Unable to decode response from API' );
        }

        return $result;
    }


    /**
     * @param string $path
     * @param array $params
     *
     * @return array|bool|float|int|object|string|null
     * @throws HttpException
     */
    protected function callEntityService( string $path, array $params = [] ) {
        $extraParams = array_merge( [
            'start'                => 0,
            'limit'                => 100,
            'order'                => 'ASC',
            'populateParentEntity' => true,
            'status'               => 'ACTIVE'
        ], $params );


        $result = $this->simpleCurl
            ->setOpt( CURLOPT_URL,
                $this->region->getEntityServicePortalUrl( $path . '?' . http_build_query( $extraParams ) ) )
            ->setOpt( CURLOPT_RETURNTRANSFER,
                true )
            ->setOpt( CURLOPT_HTTPHEADER,
                $this->getHeaders() );

        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    /**
     * @return array|bool|float|int|object|string|null
     * @throws HttpException
     */
    public function getEntities() {

        return $this->callEntityService( 'entities' );
    }

    /**
     * @param string $entityId
     *
     * @return float|object|array|bool|int|string|null
     * @throws HttpException
     */
    public function getPPCs( string $entityId ) {

        return $this->callEntityService( 'paymentContracts', [ 'parentIds' => $entityId ] );
    }

    /**
     * @return float|object|array|bool|int|string|null
     * @throws HttpException
     */
    public function get3DS() {

        return $this->callEntityService( 'threeDSContracts' );
    }

    /**
     * @return float|object|array|bool|int|string|null
     * @throws HttpException
     */
    public function getWallets() {

        return $this->callEntityService( 'walletContracts' );
    }

    public function checkCredentials(): bool {
        try {
            $result = $this->simpleCurl
                ->setOpt( CURLOPT_URL, $this->region->getTransactionCountURL() )
                ->setOpt( CURLOPT_RETURNTRANSFER, true )
                ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
            $this->addSsl( $result );

            $result = $this->decodeResult( $result->getResult() );

            return isset($result['count']);
        } catch ( Exception $exception ) {
            return false;
        }
    }

    public function getWebhooks() {
        $path   = '?eventTypes=CheckoutTransactionSuccess&order=DESC&orderBy=createdDate&pageNumber=1&pageSize=20&status=ACTIVE';
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getWebhooksUrl( $path ) )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

    public function setWebhook( $entityId, $url ) {
        $params = [
            'name'               => get_bloginfo( 'name' ),
            'eventTypes'         => [ 'CheckoutTransactionSuccess' ],
            'entities'           => [ $entityId ],
            'emailNotifications' => [],
            'endpoints'          => [ [ 'url' => $url ] ]
        ];
        $result = $this->simpleCurl->setOpt( CURLOPT_URL, $this->region->getWebhooksUrl() )
                                   ->setOpt( CURLOPT_RETURNTRANSFER, true )
                                   ->setOpt( CURLOPT_POST, true )
                                   ->setOpt( CURLOPT_POSTFIELDS, json_encode( $params ) )
                                   ->setOpt( CURLOPT_HTTPHEADER, $this->getHeaders() );
        $this->addSsl( $result );

        return $this->decodeResult( $result->getResult() );
    }

}