jQuery(function ($) {
    'use strict';
    initAutoSetup();

    $('body').on('click', '#autoSetup', function (e) {
        e.preventDefault();
        showModal();
    })

    function showModal() {
        $('.verifoneOverlay').show();
        $('body').addClass('verifoneOverlayBody');
        $('.verifoneModal').show();
    }

    function hideModal() {
        $('.verifoneOverlay').hide();
        $('body').removeClass('verifoneOverlayBody');
        $('.verifoneModal').hide();
    }

    function initAutoSetup() {

        // add this html only on verifone account settings tab
        if ($('#woocommerce_verifone_hosted_cart_account_settings').length) {
            let tr = '<tr><th class="lessPadding"><button id="autoSetup" class="button-primary">Auto setup</button></th></tr>'
            let modalHtml = '  <form><div class="card-content">' +
                '<p class="title">Auto Setup Tool</p><button class="modal-close is-large" aria-label="close">X</button>' +
                '<p>This tool will attempt to configure your Verifone account settings. Please input the necessary information below.' +
                '    <a href="https://verifone.cloud/docs/online-payments/getting-started/api-authentication/obtaining-api-keys" target="_blank">Obtain API Keys</a>' +
                '</p><hr>' +
                '<div class="field has-addons env">' +
                '    <div class="control"><button type="button" class="button" data-v="production"><span>Production</span> </button></div>' +
                '    <div class="control"><button type="button" class="button is-active" data-v="cst"><span>Sandbox</span></button></div>' +
                '</div>' +
                '<input type="hidden" id="env" name="env" value="cst">' +
                '<div class="field has-addons help has-text-danger" id="verifoneError" style="display: none">' +
                '    <span class="help has-text-danger">&nbsp;</span>' +
                '</div>' +
                '<div class="field">' +
                '    <label class="label" for="as_user_id">User ID</label>' +
                '    <div class="control">' +
                '        <input id="as_user_id" name="as_user_id" class="input" type="text" placeholder="b9b3bd52-c266-45d3-8c69-66acb7fff934">' +
                '    </div>' +
                '</div>' +
                '<div class="field">' +
                '    <label class="label" for="as_api_key">API Key</label>' +
                '    <div class="control">' +
                '        <input id="as_api_key" name="as_api_key" class="input" type="text" placeholder="mtBtnFCMYMroBWJNtloYrPSimoGXJgbiHpEc">' +
                '    </div>' +
                '</div>' +
                '<div class="field">' +
                '<label class="label" for="region">Region</label>' +
                '    <div class="select is-block">' +
                '        <select name="region" class="w-100" id="region">' +
                '            <option value="EMEA">EMEA (Europe, Middle East & Africa)</option>' +
                '            <option value="NA">NA (North America)</option>' +
                '            <option value="NZAU">NZ (New Zealand)</option>' +
                '            <option value="AU">AU (Australia)</option>' +
                '        </select>' +
                '    </div>' +
                '</div>' +
                '<div class="level pt-3">' +
                '    <button class="level-left button is-text is-primary closeModal">Cancel</button>' +
                '    <div class="buttons"><button type="button" class="button is-primary" id="startSetup">Continue</button></div>' +
                '</div>' +
                '<p class="has-text-right has-text-danger">' +
                '    <span class="woocommerce-help-tip" tabindex="0"></span>' +
                '    Clicking <span class="has-text-primary">Continue</span> will override your current settings' +
                '</p>' +
                '</div></form>';

            $('.form-table tbody').prepend(tr);
            $('body').append('<div class="verifoneOverlay"></div>')
            $('body').append('<div class="verifoneModal" id="autoSetupModal">' + modalHtml + '</div>')

            if (initialSetup === 'yes') {
                showModal();
            }

        }
    }

    // toggle env on clicking the buttons
    $('body').on('click', '.env button', function () {
        $('.env button').removeClass('is-active');
        $('#env').val($(this).data('v'));
        $(this).addClass('is-active')
    });

    $('body').on('click', '.closeModal, .modal-close', function (e) {
        e.preventDefault();
        hideModal();
    });

    $('body').on('click', '#startSetup', function (e) {
        e.preventDefault();
        let self = $(this),
            form = self.closest('form'),
            postData = form.serialize();
        $('.is-danger').removeClass('is-danger');
        $('.has-text-danger').remove();

        if ($('#as_api_key').val().length < 20) {
            $('#as_api_key').addClass('ís-danger')
            let errorHtml = '<span class="has-text-danger" style="margin: 10px 5px">Api Key must be a valid Verifone key</span>';
            $('#as_api_key').closest('.field').append(errorHtml)
            return false;
        }
        if ($('#as_user_id').val().length != 36) {
            $('#as_api_key').addClass('ís-danger');
            let errorHtml = '<span class="has-text-danger" style="margin: 10px 5px">User Id must be a valid UUID</span>';
            $('#as_user_id').closest('.field').append(errorHtml)
            return false;
        }

        if (self.hasClass('is-loading')) {
            return false;
        }

        $.ajax({
            url: VERIFONE_AUTO_SETUP_URL,
            method: 'POST',
            dataType: 'json',
            cache: false,
            data: postData,
            beforeSend: function (xhr) {
                self.addClass('is-loading');
                self.html('Processing');
            }
        }).done(function (response) {
            $('#woocommerce_verifone_hosted_cart_user_id').val(response.config.user_id);
            $('#woocommerce_verifone_hosted_cart_entity_id').val(response.config.entity_id);
            $('#woocommerce_verifone_hosted_cart_api_key').val(response.config.api_key);
            $('#woocommerce_verifone_hosted_cart_region').val(response.config.region).change();
            alert(response.message);
            $('.verifoneOverlay').hide();
            $('body').removeClass('verifoneOverlayBody');
            $('.verifoneModal').hide();
            window.onbeforeunload = '';
            setTimeout(function () {
                location.reload();
            }, 500)
        }).fail(function (response) {
            console.error(response);
            let errorMessage = "We encountered a communication error on your website and cannot proceed with the auto-setup process. Please complete your account configuration manually. Additionally, you may want to review your Wordpress installation and try again."
            if (typeof response.responseJSON !== 'undefined') {
                errorMessage = response.responseJSON.message
            }
            let error = '<span class="help has-text-danger" style="margin: 15px 0; display: block">' + errorMessage + '</span>';
            $('.env').after(error);
        }).always(function () {
            self.removeClass('is-loading');
            self.html('Continue');
        })
    });
});
