(function () {
    "use strict";

    const decodeEntities = wp.htmlEntities.decodeEntities;
    const getSetting = wc.wcSettings.getSetting;
    const registerPaymentMethod = wc.wcBlocksRegistry.registerPaymentMethod;
    const applyFilters = wp.hooks.applyFilters;
    const createElement = wp.element.createElement;

    let verifoneHostedData = getSetting("verifone_hosted_data", {});

    const Content = function (props) {
        return createElement("div", null, props.text);
    };

    const PaymentMethodLabel = function (props) {
        return createElement(props.components.PaymentMethodLabel, {className: "vhc-block-label", text: props.text});
    };

    if (verifoneHostedData.hosted.enabled)
        registerPaymentMethod({
            name: "verifone_hosted_cart",
            label: createElement(PaymentMethodLabel, {text: verifoneHostedData.hosted.title}),
            content: createElement('div', {
                dangerouslySetInnerHTML: {__html: verifoneHostedData.hosted.description}
            }),
            edit: createElement('div', {
                dangerouslySetInnerHTML: {__html: verifoneHostedData.hosted.description}
            }),
            placeOrderButtonLabel: verifoneHostedData.hosted.checkout_button_text,
            canMakePayment: function () {
                return applyFilters("vhc_blocks_enabled", true, verifoneHostedData);
            },
            ariaLabel: verifoneHostedData.hosted.title,
        });

    if (verifoneHostedData.inline.enabled)
        registerPaymentMethod({
            name: "verifone_inline_cart",
            label: createElement(PaymentMethodLabel, {text: verifoneHostedData.inline.title}),
            content: createElement('div', {
                dangerouslySetInnerHTML: {__html: verifoneHostedData.inline.description}
            }),
            edit: createElement('div', {
                dangerouslySetInnerHTML: {__html: verifoneHostedData.inline.description}
            }),
            placeOrderButtonLabel: verifoneHostedData.inline.checkout_button_text,
            canMakePayment: function () {
                return applyFilters("vhc_blocks_enabled", true, verifoneHostedData);
            },
            ariaLabel: verifoneHostedData.inline.title,
        });
})();
