<?php

use VerifoneEcomAPI\ApiWrapper\Authentication\BasicAuth;
use VerifoneEcomAPI\ApiWrapper\Http\ApiClient\Client;
use VerifoneEcomAPI\ApiWrapper\Http\AutoSetupService;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Regions\Emea;
use VerifoneEcomAPI\ApiWrapper\Regions\Na;
use VerifoneEcomAPI\ApiWrapper\Regions\NzAu;
use VerifoneEcomAPI\ApiWrapper\Regions\Au;
use VerifoneEcomAPI\ApiWrapper\Settings;

/**
 * Class WC_Verifone_Hosted_Gateway
 */
class WC_Verifone_Hosted_Gateway extends WC_Verifone {
    static $PLUGIN_ID = VERIFONE_PAYMENT_METHOD_HOSTED;

    /**
     * @inheritDoc
     */
    public function __construct() {
        parent::__construct();

        $payment_button_description = $this->get_option( 'checkout_button_text' );
        if ( empty( $payment_button_description ) ) {
            $payment_button_description = 'Proceed to Verifone Hosted Cart';
        }
        $this->order_button_text = __( $payment_button_description, 'verifone-hosted' );
        $this->verifone_settings = new Settings();
        $enable3Ds               = $this->trueOrFalse( '_3ds_enable' );
        // if the region is New Zealand we force 3ds for card payments
        if ( $this->get_option( 'region' ) === 'NZAU' ) {
            $enable3Ds = true;
        }

        $this->verifone_settings
            ->setSendTaxRates( $this->trueOrFalse( 'send_tax_rates' ) )
            ->setIsInvoice4UEnabled( $this->trueOrFalse( 'invoice4u_enabled' ) )
            ->setInvoice4UReceiptType( $this->get_option( 'invoice4u_receipt_type' ) )
            ->setTest( $this->trueOrFalse( 'test' ) )
            ->setRegion( $this->get_option( 'region' ) )
            ->setApiUserId( $this->get_option( 'user_id' ) )
            ->setApiKey( $this->get_option( 'api_key' ) )
            ->set3dsEnable( $enable3Ds )
            ->setOverlayOnPlaceOrder( $this->trueOrFalse( 'overlay_on_place_order' ) )
            ->setManageOrders( $this->trueOrFalse( 'manage_orders' ) )
            ->set3dsContractId( $this->get_option( '_3ds_contract_id' ) )
            ->setIsCardEnabled( $this->trueOrFalse( 'card_enabled' ) )
            ->setPaymentContractId( $this->get_option( 'payment_contract_id' ) )
            ->setEntityId( $this->get_option( 'entity_id' ) )
            ->setCurlVerifySsl( $this->trueOrFalse( 'curl_verify_ssl' ) )
            ->setTransactionType( $this->get_option( 'transaction_type' ) )
            ->setThemeId( $this->get_option( 'theme_id' ) )
            ->setCompleteOrderOnPayment( $this->trueOrFalse( 'complete_order_on_payment' ) )
            ->setPaypalContractId( $this->get_option( 'paypal_contract_id' ) )
            ->setIsApplePayEnabled( $this->trueOrFalse( 'apple_pay_enabled' ) )
            ->setApplePayScaComplianceLevel( $this->get_option( 'apple_pay_sca_compliance_level' ) )
            ->setApplePayContractId( $this->get_option( 'apple_pay_contract_id' ) )
            ->setIsGooglePayEnabled( $this->trueOrFalse( 'google_pay_enabled' ) )
            ->setGooglePayScaComplianceLevel( $this->get_option( 'google_pay_sca_compliance_level' ) )
            ->setGooglePayContractId( $this->get_option( 'google_pay_contract_id' ) )
            ->setIsPaypalEnabled( $this->trueOrFalse( 'paypal_enabled' ) )
            ->setIsSwishPayEnabled( $this->trueOrFalse( 'swish_pay_enabled' ) )
            ->setIsAbsInstallmentEnabled( $this->trueOrFalse( 'abs_installment_enabled' ) )
            ->setIsKlarnaEnabled( $this->trueOrFalse( 'klarna_enabled' ) )
            ->setIsGpp2Enabled( $this->trueOrFalse( 'gpp2_enabled' ) )
            ->setGpp2ContractId( $this->get_option( 'gpp2_contract_id' ) )
            ->setIsSubscriptionEnabled( $this->trueOrFalse( 'subscription_enabled' ) )
            ->setSubscriptionId( $this->get_option( 'subscription_id' ) )
            ->setIsCardFraudCheckEnabled( $this->trueOrFalse( 'fraud_check_enabled' ) )
            ->setFraudCheckContractId( $this->get_option( 'fraud_check_contract_id' ) )
            ->setIsAbsInstallmentEnabled( $this->trueOrFalse( 'abs_installment_enabled' ) )
            ->setAbsInstallmentThreshholds( $this->get_option( 'abs_installment_threshholds' ) )
            ->setIsMobilePayEnabled( $this->trueOrFalse( 'mobile_pay_enabled' ) )
            ->setMobilePayScaComplianceLevel( $this->get_option( 'mobile_pay_sca_compliance_level' ) )
            ->setMobilePayContractId( $this->get_option( 'mobile_pay_contract_id' ) )
            ->setIsBankOpEnabled( $this->trueOrFalse( 'bank_op_enabled' ) )
            ->setBankOpContractId( $this->get_option( 'bank_op_contract_id' ) )
            ->setIsVippsEnabled( $this->trueOrFalse( 'vipps_enabled' ) )
            ->setVippsScaComplianceLevel( $this->get_option( 'vipps_sca_compliance_level' ) )
            ->setVippsContractId( $this->get_option( 'vipps_contract_id' ) )
            ->setIsGiftCardEnabled( $this->trueOrFalse( 'gift_card_enabled' ) )
            ->setGiftCardContractId( $this->get_option( 'gift_card_contract_id' ) )
            ->setIsPlccEnabled( $this->trueOrFalse( 'plcc_enabled' ) )
            ->setPlccContractId( $this->get_option( 'plcc_contract_id' ) )
            ->setOoverrideRedirectUrl( $this->get_option( 'override_redirect_url' ) )
            ->setPaymentMethodsOrder( $this->get_option( 'payment_options_order' ) )
            ->setPaymentType(
                Settings::PAYMENT_TYPE_HPP == $this->get_option( 'payment_type' )
                    ? Settings::PAYMENT_TYPE_HPP
                    : Settings::PAYMENT_TYPE_IFRAME
            );

        switch ( $this->get_option( 'region' ) ) {
            case 'NZAU':
                $this->api_url_mapper = new NzAu( $this->verifone_settings );
                break;
            case 'NA':
                $this->api_url_mapper = new Na( $this->verifone_settings );
                break;
            case 'EMEA':
                $this->api_url_mapper = new Emea( $this->verifone_settings );
                break;
            case 'AU':
                $this->api_url_mapper = new Au( $this->verifone_settings );
                break;
            default:
                // will happen only on the first installation
                $this->api_url_mapper = new Emea( $this->verifone_settings );
        }

        $this->verifone_auth_helper = new BasicAuth( $this->verifone_settings );

        $this->client = new Client( $this->api_url_mapper, $this->verifone_settings, $this->verifone_auth_helper, new SimpleCurl() );
        $this->add_actions();
    }

    /**
     * Adds various actions to be used later on
     * Mostly hooks
     */
    private function add_actions() {
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [ $this, 'process_admin_options', ] );
        add_action( 'woocommerce_receipt_' . $this->id, [ $this, 'receipt_page' ] );
        add_action( 'wp_enqueue_scripts', [ $this, 'enqueue_script' ] );
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [ $this, 'reset_token' ] );
        add_action( 'woocommerce_api_payment_response', [ $this, 'verifone_check_payment_response' ] );
        add_action( 'woocommerce_api_manual_capture_payment', [ $this, 'manual_capture_payment' ] );
        add_action( 'woocommerce_order_status_cancelled', [ $this, 'cancel_payment' ] );
        add_action( 'woocommerce_order_status_completed', [ $this, 'complete_payment' ] );
        add_action( 'woocommerce_order_status_refunded', [ $this, 'cancel_payment' ] );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'verifone_checkout_capture' ], 10, 2 );
        add_action( 'admin_enqueue_scripts', array( $this, 'admin_scripts' ) );
        add_action( 'woocommerce_api_woo-verifone-webhook', [ $this, 'process_webhook' ] );
        add_action( 'woocommerce_api_woo-verifone-iframe', [ $this, 'process_iframe_checkout' ] );
        add_action( 'woocommerce_get_sections_checkout', [ $this, 'add_subsection' ] );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'add_void_transaction_button' ], 10, 1 );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'hide_order_refund_button' ] );
        add_action( 'woocommerce_api_cancel_payment', [ $this, 'manual_cancel_payment' ] );
        add_action( 'woocommerce_scheduled_subscription_payment_' . $this->id, [ $this, 'scheduled_subscription_payment' ], 10, 2 );
        add_action( 'woocommerce_api_verifone-auto-setup', [ $this, 'auto_setup' ], 10, 2 );

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            add_action( 'woocommerce_api_verifone_iframe_handle_payment_request', [ $this, 'handle_payment_request' ] );
        }
    }

    /**
     * @param $order_id
     * @param string $type
     * @param null $amount
     *
     * @return void
     */
    protected function voidOrCapture( $order_id, string $type = 'capture', $amount = null ): void {
        // we make this check if the order is updated by changing the status and if 'manageOrders' is set to TRUE,
        // if we press the 'capture amount' button, we don't make the check
        $calledBy = debug_backtrace()[1]['function'];
        if ( in_array( $calledBy, [ 'complete_payment', 'cancel_payment' ] ) && ! $this->verifone_settings->getManageOrders() ) {
            return;
        }
        $order          = wc_get_order( $order_id );
        $payment_method = $order->get_meta( VERIFONE_PAYMENT_METHOD_USED );

        if ( VERIFONE_PAYMENT_METHOD_HOSTED === $order->get_payment_method() ) {
            if ( ! $order->get_meta( VERIFONE_CAPTURE_MANUAL ) ) {
                return;
            }

            if ( $order->get_meta( VERIFONE_CAPTURE_MANUAL ) && $order->get_meta( VERIFONE_CAPTURE_MANUAL ) !== 'yes' ) {
                return;
            }
            if ( $order->get_meta( VERIFONE_CAPTURE_DONE ) && $order->get_meta( VERIFONE_CAPTURE_DONE ) === 'yes' ) {
                return;
            }

            try {
                $transaction_id = $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID );
                if ( ! $amount ) {
                    $amount = $this->calculateAmount( $order->get_total() );
                }
                $payload = [ 'amount' => $amount ];

                if ( $type == 'void' ) {
                    $result = $this->client->postVoidAuthorization( $payload, $transaction_id );
                }
                if ( $type == 'capture' ) {
                    if ( VERIFONE_PAYMENT_METHOD_PAYPAL === $payment_method ) {
                        $payload['amount'] = [
                            'value'        => $payload['amount'],
                            'currencyCode' => $order->get_currency(),
                        ];
                        $result            = $this->client->postPaypalCapture( $payload, $transaction_id );
                    } else {
                        $result = $this->client->postCapture( $payload, $transaction_id );
                    }
                }

                // a 400 request means the capture was already done
                // thus we can ignore it
                if ( isset( $result['code'] ) && $result['code'] == 400 ) {
                    return;
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_AUTHORIZED ) {
                    $message = sprintf( 'Order with transaction ID %s was authorized. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_SETTLEMENT_COMPLETED ) {
                    $message = sprintf( 'Order with transaction ID %s was settled. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_VOIDED ) {
                    $message = sprintf( 'Order with transaction ID %s was voided. Void ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && in_array( $result['status'], VERIFONE_ENV_CHECKOUT_EVENT_FAILED ) ) {
                    $message = __( "Payment was declined" );
                    $order->update_status( 'failed' );
                }

                if ( isset( $message ) ) {
                    $order->add_order_note( $message, false );
                    $order->update_meta_data( '__verifone_order_capture_done', 'yes' );
                } else {
                    $order->add_order_note( 'No change detected for order, please try setting the status again', false );
                }
                if ( is_callable( [ $order, 'save' ] ) ) {
                    $order->save();
                }
            } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
                // update the status to failed
                $order->update_status( 'failed', 'Unable to change order status, please try again later' );
                $this->logger->debug( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s . The payload: %s',
                    json_last_error(), __FILE__, __LINE__, $exception->getMessage(), json_encode( $payload ) ) );
            }
        }
    }

    /**
     * Triggered when a new payment is done for a subscription
     *
     * @param $renewal_total
     * @param WC_Order $renewal_order
     *
     * @return bool
     * @throws Exception
     */
    public function scheduled_subscription_payment( $renewal_total, WC_Order $renewal_order ) {

        $subscriptionId = null;

        foreach ( $renewal_order->get_meta_data() as $val ) {
            $arr = $val->get_data();
            if ( $arr['key'] === '_subscription_renewal' ) {
                $subscriptionId = $arr['value'];
                break;
            }
        }

        $subscription = wcs_get_subscription( $subscriptionId );
        $reference    = $subscription->get_meta( VERIFONE_ORDER_STORED_CREDENTIAL_REFERENCE );

        if ( ! $reference ) {
            $renewal_order->update_status( 'failed', 'Unable to charge card, subscription reference not found' );
            $this->logger->debug( 'The subscription reference for order ' . $renewal_order->get_id() . ' could not be found!' );
            throw new Exception( 'The subscription reference for order ' . $renewal_order->get_id() . ' could not be found!' );
        }

        $payload = [
            'amount'                    => $this->calculateAmount( $renewal_total ),
            'currency_code'             => get_option( 'woocommerce_currency' ),
            'merchant_reference'        => (string) $renewal_order->get_id(),
            'payment_provider_contract' => $this->verifone_settings->getPaymentContractId(),
            'dynamic_descriptor'        => (string) $renewal_order->get_id(),
            'stored_credential'         => [
                'stored_credential_type' => 'CHARGE',
                'processing_model'       => 'RECURRING',
                'reference'              => $reference
            ]
        ];
        try {
            $result = $this->client->postSubscription( $payload );
            $renewal_order->add_order_note( __( 'Transaction ID: ' . $result['id'] ) );
            if ( isset( $result['status'] ) && in_array( $result['status'], [ 'AUTHORIZED', 'SETTLEMENT_COMPLETED' ] ) ) {
                WC_Subscriptions_Manager::process_subscription_payments_on_order( $renewal_order );
                $renewal_order->update_status( 'completed', 'Payment success for subscription ID: ' . $subscriptionId );
            } else {
                WC_Subscriptions_Manager::process_subscription_payment_failure_on_order( $renewal_order );
                $renewal_order->update_status( 'failed', 'Payment failed for subscription ID: ' . $subscriptionId );
            }
        } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
            // update the status to failed
            $renewal_order->update_status( 'failed', 'Payment failed for subscription ID: ' . $subscriptionId );
            $this->logger->debug( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
                json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

            return false;
        }

        return true;
    }


    /**
     * Simplify configuration
     * @return void
     * @throws \VerifoneEcomAPI\ApiWrapper\Http\HttpException
     */
    public function auto_setup() {
        $params = $_POST;
        if ( empty( $params['as_user_id'] ) || empty( $params['env'] ) || empty( $params['as_api_key'] ) ) {
            wp_send_json( [ 'status' => false, 'messages' => 'UserId and/or ApiKey is missing!' ], 412 );
        }

        $service = new AutoSetupService( $params, $this->verifone_settings );
        $result  = $service->setAccount();

        wp_send_json( $result, $result['status'] ? 200 : 412 );
    }

    /**
     * ajax callable wrapper for process_payment
     *
     * @access public
     * @return void
     * @throws Exception
     */
    public function handle_payment_request() {
        ob_start();

        if ( ! empty( $_POST['terms-field'] ) && empty( $_POST['terms'] ) ) {
            wp_send_json( [
                'result'   => 'failure',
                'messages' => 'Please read and accept the terms and conditions to proceed with your order.',
                'step'     => 'existing_order',
            ] );
        }

        if ( $_SERVER['REQUEST_METHOD'] !== 'POST' || ! isset( $_POST['woocommerce-pay-nonce'] ) || ! isset( $_POST['order_id'] ) ) {
            return;
        }

        $nonce_value = $_POST['woocommerce-pay-nonce'];

        if ( ! wp_verify_nonce( $nonce_value, 'woocommerce-pay' ) ) {
            return;
        }

        $response = $this->process_payment( $_POST['order_id'] );

        wp_send_json( $response );
    }

    /**
     * Resets the token on ANY setting change
     */
    public function reset_token() {
        $token = get_option( '__verifone_api_token' );
        if ( false !== $token ) {
            delete_option( '__verifone_api_token' );
        }
    }

    /**
     * Includes a javascript file to be used in the frontend
     */
    public function enqueue_script() {
        if ( $this->verifone_settings->getOverlayOnPlaceOrder() ) {
            wp_enqueue_script( 'verifone_hosted_cart_script', plugins_url( 'assets/js/verifone_hosted_cart.js', dirname( __FILE__ ) ) );
        }

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            wp_enqueue_script( 'verifone_hosted_iframe_script', plugins_url( 'assets/js/verifone_hosted_cart_iframe.js', dirname( __FILE__ ) ), [], false, true );
            echo '<script type="text/javascript">
                const VERIFONE_IFRAME_CHECKOUT_URL="' . WC()->api_request_url( 'woo-verifone-iframe' ) . '";
                const VERIFONE_CREATING_CHECKOUT = "' . __( 'Creating the checkout, please wait....' ) . '";
            </script>';
        }
    }

    /**
     * @inheritDoc
     */
    public function admin_options() {
        require_once plugin_dir_path( __FILE__ ) . '../templates/admin-options.php';
    }

    /**
     * Includes a javascript file to be used in the backend
     */
    public function admin_scripts() {
        wp_enqueue_script( 'verifone_admin_script', plugins_url( 'assets/js/verifone_admin.js', dirname( __FILE__ ) ) );
        wp_enqueue_script( 'verifone_admin_auto_setup_script', plugins_url( 'assets/js/auto_setup.js', dirname( __FILE__ ) ), [ 'jquery' ] );
        wp_enqueue_style( 'verifone_admin_auto_setup_style', plugins_url( 'assets/css/auto_setup.css', dirname( __FILE__ ) ) );

        // check if is the first initial setup
        $initialSetup = $this->verifone_settings->getApiUserId() ? 'no' : 'yes';
        echo "<script type=\"text/javascript\">
            var verifonePaymentMethods = " . json_encode( $this->verifone_settings->getPaymentMethods() ) . ";
            const VERIFONE_AUTO_SETUP_URL='" . WC()->api_request_url( 'verifone-auto-setup' ) . "';
            const initialSetup = '" . $initialSetup . "'; 
        </script>";
    }

    /**
     * @inheritDoc
     */
    public function init_form_fields() {
        require_once plugin_dir_path( __FILE__ ) . '../templates/init_form_fields.php';
        $current_subsection = ( isset( $_GET['vhc-tab'] ) && ! empty( $_GET['vhc-tab'] ) ) ? $_GET['vhc-tab'] : '';

        switch ( $current_subsection ) {
            case 'verifone-account-settings-tab':
            default:
                $this->form_fields = getVerifoneHostedCartVerifoneAccountSettingsFields();
                break;
            case 'card-payment-settings-tab':
                $this->form_fields = getVerifoneHostedCartCardPaymentSettingsFields();
                break;
            case 'alternative-payment-options-tab':
                $this->form_fields = getVerifoneHostedCartAlternativePaymentOptionsFields();
                break;
            case 'display-settings-tab':
                $this->form_fields = getVerifoneHostedCartDisplaySettingsFields();
                break;
            case 'advanced-settings-tab':
                $this->form_fields = getVerifoneHostedCartAdvancedSettingsFields();
                break;
        }
    }

    public function process_iframe_checkout() {
        if ( $_SERVER['REQUEST_METHOD'] !== 'POST' || $_POST['payment_method'] !== self::$PLUGIN_ID ) {
            $this->api_response_error();
        }

        $cart     = WC()->cart;
        $customer = WC()->customer;
        $order    = new WC_Order();

        foreach ( $cart->get_cart() as $cart_item_key => $cart_item ) {
            $product_id   = $cart_item['product_id'];
            $variation_id = $cart_item['variation_id'];
            $quantity     = $cart_item['quantity'];

            // Add line items to the order
            $order->add_product( wc_get_product( $product_id ), $quantity, array(
                'variation_id' => $variation_id
            ) );
        }

        $coupons = $cart->get_coupons();
        if ( $coupons ) {
            foreach ( $coupons as $coupon ) {
                $order->apply_coupon( $coupon );
            }
        }

        $order->set_address( $customer->get_billing(), 'billing' );
        $order->set_address( $customer->get_shipping(), 'shipping' );

        $order->calculate_totals();
        $order->save();

        return $this->api_response_ok( $this->process_payment( $order->get_id() ) );
    }

    /**
     * @param int $order_id
     *
     * @return array|\WP_Error
     * @throws Exception
     */
    public function process_payment( $order_id ) {
        if ( $this->verifone_settings->is3dsEnabled() && null === $this->verifone_settings->get3dsContractId() ) {
            $this->logger->debug( '3DS is enabled but no 3DS contract ID is provided in the configuration. Please fill in the field "3DS Contract Id" or disable the field "Enable/Disable 3d-secure" in the plugin settings.' );

            return $this->failure();
        }

        $order = wc_get_order( $order_id );
        if ( ! $order instanceof WC_Order ) {
            return $this->failure();
        }

        $payment_gateways = WC()->payment_gateways->payment_gateways();
        if ( ! isset( $payment_gateways[ self::$PLUGIN_ID ] ) ) {
            $this->logger->debug( sprintf( 'Attempted to set payment method as %s however said payment method does not exist. This error should not happen unless the payment method is misspelled.',
                self::$PLUGIN_ID ) );

            return $this->failure();
        }

        $order->set_payment_method( $payment_gateways[ self::$PLUGIN_ID ] );
        $order->save();
        $locale = $this->s40( get_locale(), 2 );
        if ( in_array( $locale, self::NORWAY_CODES ) ) {
            $locale = "no";
        }

        //create the customer - will return NULL if the customer could not be created
        $customer = $this->getCustomer( $order );
        if ( ! $customer ) {
            return $this->failure();
        }

        $transaction_mode = wp_is_mobile() ? 'P' : 'S';
        $orderAmount      = $this->calculateAmount( $order->get_total() );
        $order->update_meta_data( VERIFONE_CAPTURE_MANUAL, $this->verifone_settings->isManualCapture() ? 'yes' : 'no' );
        $order->update_meta_data( VERIFONE_ORDER_TRANSACTION_TYPE, $this->verifone_settings->getAuthType() );
        $order->save_meta_data();
        $redirectUrl = add_query_arg( [ 'wc-api' => 'payment_response', 'pm' => 'verifone_hosted', ], home_url( '/' ) );

        $checkout = [
            'entity_id'          => $this->verifone_settings->getEntityId(),
            'currency_code'      => get_option( 'woocommerce_currency' ),
            'amount'             => $orderAmount,
            'return_url'         => $redirectUrl,
            'customer'           => $customer,
            'redirect_method'    => 'HEADER_REDIRECT',
            'shop_url'           => wc_get_checkout_url(),
            'merchant_reference' => (string) $order->get_id(),
            'configurations'     => [],
            'i18n'               => [
                'default_language'  => $locale,
                'fallback_language' => 'en'
            ],
        ];
        // we use this as a temp fix for Klarna
        $customerLocale = $order->get_billing_country();

        $this->card( $checkout, $transaction_mode );
        $this->bankOp( $checkout, $order );
        $this->mobilPay( $checkout, $order, $transaction_mode );
        $this->paypal( $checkout );
        $this->klarna( $checkout, $orderAmount, $customerLocale );
        $this->googlePay( $checkout, $transaction_mode );
        $this->swish( $checkout, $transaction_mode );
        $this->applePay( $checkout, $transaction_mode );
        $this->vipps( $checkout, $order, $transaction_mode );
        $this->gpp2( $checkout );
        $this->subscriptions( $checkout, $order );
        $this->installments( $checkout, $order );
        $this->giftCard( $checkout );
        $this->plcc( $checkout );
        $this->invoice4u( $checkout, $order );
        $this->theme( $checkout );

        $this->orderPaymentMethods( $checkout );
        try {
            $result = $this->client->postCheckout( $checkout );
        } catch ( Exception $exception ) {
            $this->logger->debug(
                'There was an error in the API response: ' . PHP_EOL .
                'Sent request: ' . json_encode( $checkout ) . PHP_EOL .
                'Received response: ' . $exception->getMessage() . PHP_EOL .
                'To endpoint: ' . $this->api_url_mapper->postCheckoutUrl()
            );

            return $this->failure( 'Something went wrong!' );
        }

        if ( ! isset( $result['url'] ) ) {
            $this->logger->debug( sprintf( 'The "url" key is not present in the verifone response, unable to continue, in file %s, at line %s. API response %s',
                __FILE__, __LINE__, json_encode( $result ) ) );

            return $this->failure();
        }

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            // the host seems to be different from any url we currently use
            $scheme = parse_url( $result['url'], PHP_URL_SCHEME );
            $host   = parse_url( $result['url'], PHP_URL_HOST );

            return [
                'result'  => 'success',
                'payload' => wp_json_encode( [
                    'url' => sprintf( '%s://%s/v2/loader.js?checkoutId=%s', $scheme, $host, $result['id'] ),
                ] ),
            ];
        }

        return [
            'result'   => 'success',
            'redirect' => $result['url'],
        ];
    }

    /**
     * @throws Exception
     */
    public function verifone_check_payment_response() {
        $params = $_GET;

        if ( ! isset( $params['pm'] ) || $params['pm'] != 'verifone_hosted' ) {
            return;
        }

        if ( ! isset( $params['checkout_id'] ) ) {
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            $this->logger->debug( sprintf( 'The checkout_id parameter could not be found, in order to proceed it must exist. Current GET params: %s',
                json_encode( $_GET ) ) );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        try {
            $response = $this->client->getCheckout( $params['checkout_id'] );
        } catch ( Exception $exception ) {
            $this->logger->debug( 'There was an error in the API response: ' . PHP_EOL .
                                  'GET parameters: ' . json_encode( $params ) . PHP_EOL .
                                  'Received response: ' . $exception->getMessage() . PHP_EOL .
                                  'To endpoint: ' . $this->api_url_mapper->getCheckoutUrl( $params['checkout_id'] )
            );

            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        if ( ! isset( $response['events'] )
             || ! isset( $response['merchant_reference'] )
             || ! isset( $response['transaction_id'] )
        ) {
            $this->logger->debug( sprintf( 'The events or merchant_reference or transactionId parameter could not be found in response, in order to proceed it must exist. API response %s',
                json_encode( $response ) ) );
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        $order = wc_get_order( (int) $response['merchant_reference'] );
        if ( ! $order instanceof WC_Order ) {
            $this->logger->debug( 'There was a request for an order that doesn\'t exist in current shop! Requested params: ' . json_encode( $_GET ) );
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        // save the transaction id too
        $this->orderCheckAndLogEvent( $order, $response['id'] );
        $order->update_meta_data( VERIFONE_ORDER_TRANSACTION_ID, $response['transaction_id'] );
        $order->update_meta_data( VERIFONE_PAYMENT_METHOD_USED, $response['payment_method_used'] );
        $order->save_meta_data();

        $checkoutSuccess = false;

        // set the order status based on the checkout response
        // I was only able to get 3 statuses to get pushed back
        foreach ( $response['events'] as $event ) {
            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_SUCCESS == $event['type'] ) {

                $checkoutSuccess = true;

                if ( $order->get_meta( VERIFONE_ORDER_TRANSACTION_TYPE ) == 'SALE' && ( ! $event['details']['details']['auto_capture'] || VERIFONE_TRANSACTION_AUTHORIZED === $event['details']['status'] ) ) {
                    if ( ! in_array( $order->get_status(), [ 'completed', 'processing' ] ) ) {
                        $order->add_order_note( __( 'The payment is not settled yet. 
                        Attempting to refund the order in full, before the order is settled, will void the transaction and cancel the payment.
                        Order status changed to Pending.' ), false );

                        $order->update_status( 'processing' );
                    }
                } elseif ( $order->get_meta( VERIFONE_CAPTURE_MANUAL ) === 'yes' ) {
                    if ( ! in_array( $order->get_status(), [ 'completed', 'processing' ] ) && ! $this->noteExists( $order->get_id(), '(transaction is in PRE/FINAL AUTH status)' ) ) {
                        $order->add_order_note( __( 'The payment is not settled yet (transaction is in PRE/FINAL AUTH status), use the CAPTURE button to fully/partial capture the amount. 
                        Attempting to make a full refund, before capturing, will void the authorization and cancel the payment.
                        Order status changed to Pending.' ), false );
                    }
                    $order->update_status( 'processing' );
                } else {
                    $order->payment_complete();
                    if ( $this->verifone_settings->getCompleteOrderOnPayment() || $this->isOrderOnlyVirtual( $order ) ) {
                        //Cannot update status to completed for already active subscriptions.
                        if ( ! ( $order instanceof WC_Subscription ) || 'active' !== $order->get_status() ) {
                            $order->update_status( 'completed' );
                        }
                    }
                    $order->add_order_note( __( "Order payment is completed." ), false );
                }
                $this->hasSubscriptions( $event, $order );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_FAILED == $event['type'] ) {
                $order->add_order_note( __( "Payment was declined" ), false );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TIMEOUT_ERROR == $event['type'] ) {
                $order->add_order_note( __( "Payment was declined" ), false );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_DECLINED == $event['type'] ) {
                $order->add_order_note( __( "Payment was declined" ), false );
            }
        }

        $order->save();
        $redirect_url = $this->verifone_settings->getOverrideRedirectUrl();
        if ( ! $redirect_url || ! filter_var( $redirect_url, FILTER_VALIDATE_URL ) ) {
            $redirect_url = $order->get_checkout_order_received_url();
        }

        if ($checkoutSuccess) {
            if ( wp_redirect( $redirect_url ) ) {
                exit;
            }
        } else {
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }
    }

    public function post_capture( $order ) {

    }

    /**
     * $customerLocale is a quick fix that enables klarna only for certain countries
     *
     * @param array $checkout
     * @param int $orderAmount
     * @param string $customerLocale
     *
     * @return void
     */
    private function klarna( array &$checkout, int $orderAmount, string $customerLocale ) {
        if ( $this->verifone_settings->getIsKlarnaEnabled() &&
             in_array( $customerLocale, [ 'US', 'NO', 'SE', 'FI', 'DK' ] ) ) {
            $checkout['configurations']['klarna'] = [
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];
            // we don't need the line items if we have subscription
            if ( ! class_exists( 'WC_Subscriptions_Cart' ) ||
                 ( class_exists( 'WC_Subscriptions_Cart' ) &&
                   ! WC_Subscriptions_Cart::cart_contains_subscription() ) ) {
                $checkout['line_items'] [] = [
                    'name'             => substr( preg_replace( '/[^A-Za-z0-9\-]/', ' ', get_bloginfo( 'name' ) ), 0, 255 ),
                    "quantity"         => 1,
                    "tax_rate"         => 0,
                    "total_tax_amount" => 0,
                    "unit_price"       => $orderAmount,
                    "total_amount"     => $orderAmount,
                    "category"         => "DIGITAL_GOODS"
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function applePay( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isApplePayEnabled() ) {
            $checkout['configurations']['apple_pay'] = [
                'card'        => [
                    'sca_compliance_level' => $this->verifone_settings->getApplePayScaComplianceLevel(),
                    'payment_contract_id'  => $this->verifone_settings->getApplePayContractId(),
                ],
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];

            if ( $this->verifone_settings->getApplePayScaComplianceLevel() === 'FORCE_3DS' ) {
                $checkout['configurations']['apple_pay']['card']['threed_secure'] = [
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function googlePay( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isGooglePayEnabled() ) {

            $checkout['configurations']['google_pay'] = [
                'card'        => [
                    'sca_compliance_level' => $this->verifone_settings->getGooglePayScaComplianceLevel(),
                    'payment_contract_id'  => $this->verifone_settings->getGooglePayContractId(),
                ],
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];
            if ( in_array( $this->verifone_settings->getGooglePayScaComplianceLevel(), [ 'WALLET', 'FORCE_3DS' ] ) ) {
                $checkout['configurations']['google_pay']['card']['threed_secure'] = [
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function swish( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isSwishPayEnabled() ) {
            $checkout['configurations']['swish'] = [
                'shopper_interaction' => $transaction_mode === "P" ? 'MCOMMERCE' : 'ECOMMERCE',
            ];
        }
    }

    /**
     * @param $checkout
     *
     * @return void
     */
    private function paypal( &$checkout ) {
        if ( $this->verifone_settings->isPaypalEnabled() ) {
            $checkout['configurations']['paypal'] = [
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getPaypalContractId(),
            ];
        }
    }

    private function lineItems( array &$checkout, WC_Order $order ) {
        if ( ! empty( $checkout['line_items'] ) ) {
            return;
        }

        $all_tax_rates = [];
        $tax_classes   = WC_Tax::get_tax_classes(); // Retrieve all tax classes.
        if ( ! in_array( '', $tax_classes ) ) { // Make sure "Standard rate" (empty class name) is present.
            array_unshift( $tax_classes, '' );
        }
        foreach ( $tax_classes as $tax_class ) { // For each tax class, get all rates.
            $taxes         = WC_Tax::get_rates_for_tax_class( $tax_class );
            $all_tax_rates = array_merge( $all_tax_rates, $taxes );

        }
        foreach ( $order->get_items() as $k => $itemData ) {
            $product  = wc_get_product( $itemData->get_product_id() );
            $item     = $itemData->get_data();
            $lineItem = [
                'name'         => substr( preg_replace( '/[^A-Za-z0-9\-]/', ' ', $item['name'] ), 0, 248 ),
                "quantity"     => $item['quantity'],
                "unit_price"   => $this->calculateAmount( $order->get_item_total( $itemData, true, true ) ),
                "total_amount" => $this->calculateAmount( $order->get_line_total( $itemData, true, true ) ),
                "category"     => $product->is_virtual() ? "DIGITAL_GOODS" : "PHYSICAL_GOODS",
            ];
            if ( $this->verifone_settings->getSendTaxRates() ) {
                foreach ( $all_tax_rates as $rate ) {
                    if ( array_key_first( $item['taxes']['total'] ) == $rate->tax_rate_id ) {
                        $lineItem['tax_rate'] = (int) $rate->tax_rate;
                    }
                }
            }

            $checkout['line_items'] [] = $lineItem;
        }
        $shippingRatePercent = 0;
        foreach ( WC()->session->get( 'shipping_for_package_0' )['rates'] as $method_id => $shippingRate ) {
            if ( WC()->session->get( 'chosen_shipping_methods' )[0] == $method_id ) {
                foreach ( $all_tax_rates as $rate ) {
                    if ( array_key_first( $shippingRate->get_taxes() ) == $rate->tax_rate_id ) {
                        $shippingRatePercent = (int) $rate->tax_rate;
                        break;
                    }
                }
                break;
            }
        }
        // add shipping cost if exists
        if ( $order->get_data()['shipping_total'] > 0 ) {
            $totalAmount = $this->calculateAmount( $order->get_data()['shipping_total'] + $order->get_data()['shipping_tax'] );

            $checkout['line_items'][] = [
                'name'         => __( 'Shipping fee' ),
                "quantity"     => 1,
                "unit_price"   => $totalAmount,
                "tax_rate"     => $shippingRatePercent,
                "total_amount" => $totalAmount,
                "category"     => "DIGITAL_GOODS"
            ];
        }
    }

    private function invoice4u( &$checkout, $order ) {
        if ( $this->verifone_settings->getIsInvoice4UEnabled() ) {
            $checkout['receipt_type'] = $this->verifone_settings->getInvoice4UReceiptType();
            // we add the line item products
            $this->lineItems( $checkout, $order );

            // add discount if exists
            // we don't use it at this time
//            if ( $order->get_data()['discount_total'] > 0 ) {
//                $discount = $this->calculateAmount($order->get_data()['discount_total'] + $order->get_data()['discount_tax']);
//                $checkout['line_items'][] = [
//                    'name'         => __( 'Discount' ),
//                    "quantity"     => 1,
//                    "unit_price"   => - $discount,
//                    "total_amount" => - $discount,
//                    "tax_rate"     => 0,
//                ];
//            }
        }
    }


    /**
     * @param $checkout
     *
     * @return void
     */
    private function gpp2( &$checkout ) {
        if ( $this->verifone_settings->getIsGpp2Enabled() ) {
            $checkout['configurations']['gpp2'] = [
                'payment_contract_id' => $this->verifone_settings->getGpp2ContractId(),
            ];
        }
    }

    /**
     * @param $checkout
     * @param $order
     *
     * @return void
     */
    private function installments( &$checkout, $order ) {
        if ( $this->verifone_settings->getIsCardEnabled() && $this->verifone_settings->isAbsInstallmentEnabled() ) {
            $installment = $this->verifone_settings->getAbsInstallmentThreshholdForAmount( $order->get_total() );
            if ( $installment && $installment >= Settings::MIN_INSTALLMENT_PAYMENT_DISPLAY_NUMBER ) {
                $checkout['configurations']['card']['credit_term'] = 'INSTALMENT_STANDARD';
                $checkout['configurations']['card']['instalment']  = [
                    'max_number_of_instalments' => $installment
                ];
            } else {
                $checkout['configurations']['card']['credit_term'] = 'STANDARD';
            }
        }

    }

    /**
     * @param $checkout
     *
     * @return void
     */
    private function theme( &$checkout ) {
        if ( ! empty ( $this->verifone_settings->getThemeId() ) ) {
            $checkout['theme_id'] = $this->verifone_settings->getThemeId();
        }
    }

    private function orderPaymentMethods( &$checkout ) {
        $paymentMethods             = $checkout['configurations'];
        $checkout['configurations'] = [];

        foreach ( $this->verifone_settings->getPaymentMethodsOrder() as $i => $pmKey ) {
            if ( ! empty( $paymentMethods[ $pmKey ] ) ) {
                $checkout['configurations'][ $pmKey ] = $paymentMethods[ $pmKey ];
                unset( $paymentMethods[ $pmKey ] );
            }
        }

        //Add the rest of PMs that are not ordered but exists
        foreach ( $paymentMethods as $pmKey => $pm ) {
            $checkout['configurations'][ $pmKey ] = $pm;
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     * @param $order
     * @param string $type
     *
     * @return void
     */
    private function card( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->getIsCardEnabled() ) {
            $threeDS = [];

            $settings = [
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getPaymentContractId(),
                'authorization_type'  => $this->verifone_settings->getAuthType(),
            ];

            if ( $this->verifone_settings->getIsCardFraudCheckEnabled() ) {
                $settings['fraud_protection_contract'] = $this->verifone_settings->getFraudCheckContractId();
            }

            if ( $this->verifone_settings->is3dsEnabled() ) {
                $threeDS = [
                    'threed_secure' => [
                        'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                        'enabled'             => $this->verifone_settings->is3dsEnabled(),
                        'transaction_mode'    => $transaction_mode,
                    ]
                ];
            }

            $checkout['configurations']['card'] = array_merge( $settings, $threeDS );
        }
    }

    private function giftCard( &$checkout ) {
        if ( $this->verifone_settings->getIsGiftCardEnabled() ) {
            $checkout['configurations']['gift_card'] = [
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
                'card'        => [
                    'payment_contract_id' => $this->verifone_settings->getGiftCardContractId(),
                ]
            ];
        }
    }

    private function plcc( &$checkout ) {
        if ( $this->verifone_settings->getIsPlccEnabled() ) {
            $checkout['configurations']['plcc'] = [
                'mode'                => 'PAYMENT',
                'payment_contract_id' => $this->verifone_settings->getPlccContractId(),
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'authorization_type'  => $this->verifone_settings->getAuthType()
            ];
        }
    }

    private function subscriptions( &$checkout, $order, $type = 'SIGNUP' ) {
        if ( $this->verifone_settings->getIsCardEnabled() ) {
            //check if order has any subscription && the subscription is enabled
            if ( class_exists( 'WC_Subscriptions_Cart' ) &&
                 ( WC_Subscriptions_Cart::cart_contains_subscription() || $order instanceof WC_Subscription ) &&
                 $this->verifone_settings->getIsSubscriptionEnabled() ) {

                // unset all other payment method except CARD
                // subscriptions are supported only by CARD
                foreach ( $checkout['configurations'] as $key => $value ) {
                    if ( $key !== 'card' ) {
                        unset( $checkout['configurations'][ $key ] );
                    }
                }

                $checkout['configurations']['card']['token_preference'] = [
                    'token_scope' => $this->verifone_settings->getSubscriptionId(),
                    'token_type'  => 'REUSE'
                ];

                $checkout['configurations']['card']['stored_credential'] = [
                    'stored_credential_type'   => $type,
                    // "first_payment" --- Set to true if this signup request also is the first payment. Most relevant for signup using a minor amount,
                    // to indicate if that minor amount is just a signup account verification amount or an actual first payment amount as well.
                    // Default on this endpoint is that zero amount is not a payment,
                    // but any amount provided is a payment. Use this indicator to explicitly tell the intention
                    'first_payment'            => $order->get_total() > 0,
                    'processing_model_details' => [
                        'processing_model'  => 'RECURRING',
                        'payment_frequency' => [
                            'time_unit' => $this->getOrderBillingPeriodForSubscription( $order ),
                            'value'     => (string) $this->calculateAmount( WC_Subscriptions_Order::get_recurring_total( $order ) )
                        ]
                    ],
                ];
            }
        }
    }

    /**
     * @param $order
     *
     * @return string
     */
    private function getOrderBillingPeriodForSubscription( $order ) {
        if ( $order instanceof WC_Subscription ) {
            $billing_period = $order->get_billing_period();
        } else {
            foreach ( wcs_get_subscriptions_for_order( $order, array( 'order_type' => 'parent' ) ) as $subscription ) {

                // Find the billing period discount for all recurring items
                if ( empty( $product_id ) ) {
                    $billing_period = $subscription->get_billing_period();
                    break;
                } else {
                    // We want the billing period for a specific item (so we need to find if this subscription contains that item)
                    foreach ( $subscription->get_items() as $line_item ) {
                        if ( wcs_get_canonical_product_id( $line_item ) == $product_id ) {
                            $billing_period = $subscription->get_billing_period();
                            break 2;
                        }
                    }
                }
            }
        }

        return strtoupper( $billing_period );
    }

    /**
     * @param $checkout
     * @param $transaction_mode P - mobile, S - Computer device, T - tablet device
     * @param $order
     * @param string $type
     *
     * @return void
     */
    protected function addCardPayload( &$checkout, $transaction_mode ) {
        $threeDS = [];

        $settings = [
            'capture_now'         => ! $this->verifone_settings->isManualCapture(),
            'payment_contract_id' => $this->verifone_settings->getPaymentContractId(),
            'authorization_type'  => $this->verifone_settings->getAuthType(),
        ];

        if ( $this->verifone_settings->getIsCardFraudCheckEnabled() ) {
            $settings['fraud_protection_contract'] = $this->verifone_settings->getFraudCheckContractId();
        }

        if ( $this->verifone_settings->is3dsEnabled() ) {
            $threeDS = [
                'threed_secure' => [
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'enabled'             => $this->verifone_settings->is3dsEnabled(),
                    'transaction_mode'    => $transaction_mode,
                ]
            ];
        }

        $checkout['configurations']['card'] = array_merge( $settings, $threeDS );
    }

    private function bankOp( &$checkout ) {
        if ( $this->verifone_settings->isBankOpEnabled() && in_array( $checkout['currency_code'], BANKOP_CURRENCIES ) ) {
            $checkout['configurations']['bank'] = [
                'authorization_type' => Settings::AUTH_TYPE_FINAL_AUTH,
                'op'                 => [
                    'payment_contract_id' => $this->verifone_settings->getBankOpContractId(),
                ],
            ];
        }
    }

    /**
     * @param $checkout
     * @param $order
     * @param $transaction_mode
     *
     * @return void
     */
    private function mobilPay( &$checkout, $order, $transaction_mode ) {
        if ( $this->verifone_settings->isMobilePayEnabled() ) {
            $checkout['configurations']['mobile_pay'] = [
                "dynamic_descriptor"  => (string) $order->get_id(),
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getMobilePayContractId(),
                'card'                => [
                    "sca_compliance_level" => $this->verifone_settings->getMobilePayScaComplianceLevel(),
                    "authorization_type"   => $this->verifone_settings->getAuthType(),
                ]
            ];
            if ( $this->verifone_settings->is3dsEnabled() ) {
                $checkout['configurations']['mobile_pay']['card']['threed_secure'] = [
                    "threeds_contract_id" => $this->verifone_settings->get3dsContractId(),
                    "transaction_mode"    => $transaction_mode,
                    "enabled"             => true,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $order
     * @param $transaction_mode
     *
     * @return void
     */
    private function vipps( &$checkout, $order, $transaction_mode ) {
        if ( $this->verifone_settings->isVippsEnabled() ) {
            $checkout['configurations']['vipps'] = [
                'dynamic_descriptor'  => (string) $order->get_id(),
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getVippsContractId(),
                'card'                => [
                    'sca_compliance_level' => $this->verifone_settings->getVippsScaComplianceLevel(),
                    'authorization_type'   => $this->verifone_settings->getAuthType()
                ]
            ];

            if ( 'NONE' !== $this->verifone_settings->getVippsScaComplianceLevel() && $this->verifone_settings->get3dsContractId() ) {
                $checkout['configurations']['vipps']['card']['threed_secure'] = [
                    'enabled'             => true,
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'total_items'         => str_pad( count( $order->get_items() ), 2, '0', STR_PAD_LEFT ),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    public function add_subsection( $settings_tabs ) {
        $current_page = ( isset( $_GET['tab'] ) && ! empty( $_GET['tab'] ) ) ? $_GET['tab'] : '';
        if ( $current_page == 'checkout' && $settings_tabs !== null ) {
            $current_section = ( isset( $_GET['section'] ) && ! empty( $_GET['section'] ) ) ? $_GET['section'] : '';
            if ( $current_section == 'verifone_hosted_cart' ) {
                $sections = array(
                    'verifone-account-settings-tab'   => __( 'Verifone account settings', 'woocommerce-verifone-account-settings-tab' ),
                    'display-settings-tab'            => __( 'Display settings', 'woocommerce-display-settings-tab' ),
                    'card-payment-settings-tab'       => __( 'Card payment settings', 'woocommerce-card-payment-settings-tab' ),
                    'alternative-payment-options-tab' => __( 'Alternative payment options', 'woocommerce-alternative-payment-options-tab' ),
                    'advanced-settings-tab'           => __( 'Advanced settings', 'woocommerce-advanced-settings-tab' ),
                );
                echo '<h2 style="display: inline-block">' . $this->method_title . '</h2>';
                wc_back_link( __( 'Return to payments', 'woocommerce' ), admin_url( 'admin.php?page=wc-settings&tab=checkout' ) );
                echo '<p>' . $this->method_long_description . ' <a href="' . $this->method_description_link_url . '" target="_blank">' . $this->method_description_link_text . '</a></p>';
                echo '<p>' . __( 'Register this link to your Verifone Account', 'verifone-hosted' ) . '</p>
                    <div style="display: inline-block; padding: 5px 20px; border: 1px solid #BBBBBB; margin-bottom: 30px; border-radius: 2px; background-color: #EEEEEE;">
                        <button style="margin-right: 10px; cursor: pointer; display: inline-block; border: 0" id="webhook_copy_button" type="button">
                            <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-clipboard" viewBox="0 0 16 16">
                              <path d="M4 1.5H3a2 2 0 0 0-2 2V14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V3.5a2 2 0 0 0-2-2h-1v1h1a1 1 0 0 1 1 1V14a1 1 0 0 1-1 1H3a1 1 0 0 1-1-1V3.5a1 1 0 0 1 1-1h1v-1z"/>
                              <path d="M9.5 1a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5h-3a.5.5 0 0 1-.5-.5v-1a.5.5 0 0 1 .5-.5h3zm-3-1A1.5 1.5 0 0 0 5 1.5v1A1.5 1.5 0 0 0 6.5 4h3A1.5 1.5 0 0 0 11 2.5v-1A1.5 1.5 0 0 0 9.5 0h-3z"/>
                            </svg>
                        </button>
                    <p style="display: inline-block" id="webhook_response_url">' . $this->get_webhook_response_url() . '</p>
                    </div>';
                echo '<br/>';
                echo '<ul style="margin-bottom: 20px" class="subsubsub">';
                $array_keys         = array_keys( $sections );
                $key                = "vhc-tab";
                $current_subsection = ( isset( $_GET['vhc-tab'] ) && ! empty( $_GET['vhc-tab'] ) ) ? $_GET['vhc-tab'] : 'default';
                foreach ( $sections as $id => $label ) {
                    $url = admin_url( 'admin.php?page=wc-settings&tab=checkout&section=verifone_hosted_cart&' . $key . '=' . $id );
                    echo '<li><a href="' . esc_url( $url ) . '" class="' . ( $current_subsection === $id || ( $current_subsection === 'default' && $id === 'verifone-account-settings-tab' ) ? 'current' : '' ) . '">' . esc_html( $label ) . '</a> ' . ( end( $array_keys ) === $id ? '' : '|' ) . ' </li>';
                }
                echo '</ul><br class="clear" />';
            }
        }
    }

    public function add_void_transaction_button( $order ) {
        if ( strtolower( $order->get_data()['status'] ) === strtolower( VERIFONE_TRANSACTION_PENDING ) ) {
            $label = __( 'Cancel transaction', 'verifone-hosted' );
            $slug  = 'cancel-transaction';
            $url   = get_site_url() . '/?wc-api=cancel_payment&order_id=' . $order->get_id();
            include( plugin_dir_path( __FILE__ ) . '../templates/html-void-transaction.php' );
        }
    }

    public function manual_cancel_payment() {
        ob_start();
        $params   = $_POST;
        $order_id = $params['order_id'];
        $this->cancel_payment( $order_id );
        $order = wc_get_order( $order_id );
        $note  = 'Payment transaction cancelled.';
        $order->update_status( 'cancelled', $note );
        wp_send_json( [ 'result' => 'success', 'message' => $note ], 200 );
    }

    public function process_refund( $order_id, $amount = null, $reason = '' ) {
        $order = wc_get_order( $order_id );

        if ( VERIFONE_PAYMENT_METHOD_HOSTED !== $order->get_payment_method() ) {
            return;
        }

        return $this->_process_refund( $order_id, $amount, $reason );
    }

    public function hide_order_refund_button( $order ) {
        if ( ! isset( $order->get_data()['id'] ) ) {
            return;
        }
        if ( ! isset( $order->get_data()['date_paid'] ) ) {
            ?>
            <script>
                jQuery(function () {
                    jQuery('.refund-items').hide();
                    jQuery('.order_actions option[value=send_email_customer_refunded_order]').remove();
                    jQuery('#order_status option[value=wc-refunded]').remove();
                });
            </script>
            <?php
        }
    }

    function getPluginId(): string {
        return self::$PLUGIN_ID;
    }

    function get_webhook_response_url() {
        return WC()->api_request_url( 'woo-verifone-webhook' );
    }
}
