<?php

use VerifoneEcomAPI\ApiWrapper\Authentication\BasicAuth;
use VerifoneEcomAPI\ApiWrapper\Http\ApiClient\Client;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Regions\Emea;
use VerifoneEcomAPI\ApiWrapper\Regions\Na;
use VerifoneEcomAPI\ApiWrapper\Regions\NzAu;
use VerifoneEcomAPI\ApiWrapper\Regions\Au;
use VerifoneEcomAPI\ApiWrapper\Settings;

/**
 * Class WC_Verifone_Hosted_Gateway
 */
class WC_Verifone_Inline_Gateway extends WC_Verifone {
    static $PLUGIN_ID = VERIFONE_PAYMENT_METHOD_INLINE;
    static $SUBSECTIONS_RENDERED = false;

    public function __construct() {
        parent::__construct();

        $this->id = $this->getPluginId();
        $this->plugin_name = 'Verifone Inline Cart Payment';
        $this->title = $this->get_option('title', __('Verifone Inline Cart', 'verifone-hosted'));
        $this->method_title = __('Verifone Inline Cart', 'verifone-hosted');
        $this->method_long_description = __('Accept card payments and other payment options available in your Verifone account.', 'verifone-hosted');
        $payment_button_description = $this->get_option('checkout_button_text');
        if (empty($payment_button_description)) {
            $payment_button_description = 'Proceed to VerifoneJS Cart';
        }
        $this->order_button_text = __($payment_button_description, 'verifone-hosted');
        $this->has_fields=true;

        $this->verifone_settings = new Settings();
        $this->verifone_settings
            ->setTest($this->trueOrFalse('test'))
            ->setRegion($this->get_option('region'))
            ->setApiUserId($this->get_option('user_id'))
            ->setApiKey($this->get_option('api_key'))
            ->setIsCardEnabled($this->trueOrFalse('card_enabled'))
            ->setPaymentContractId($this->get_option('payment_contract_id'))
            ->setEntityId($this->get_option('entity_id'))
            ->setCurlVerifySsl($this->trueOrFalse('curl_verify_ssl'))
            ->setTransactionType($this->get_option('transaction_type'))
            ->setCompleteOrderOnPayment($this->trueOrFalse('complete_order_on_payment'))
            ->setIsSubscriptionEnabled($this->trueOrFalse('subscription_enabled'))
            ->setSubscriptionId($this->get_option('subscription_id'))
            ->setOoverrideRedirectUrl($this->get_option('override_redirect_url'))
            ->setManageOrders($this->trueOrFalse('manage_orders'));

        switch ($this->get_option('region')) {
            case 'NZAU':
                $this->api_url_mapper = new NzAu($this->verifone_settings);
                break;
            case 'NA':
                $this->api_url_mapper = new Na($this->verifone_settings);
                break;
            case 'EMEA':
                $this->api_url_mapper = new Emea($this->verifone_settings);
                break;
            case 'AU':
                $this->api_url_mapper = new Au($this->verifone_settings);
                break;
            default:
                // will happen only on the first installation
                $this->api_url_mapper = new Emea($this->verifone_settings);
        }

        $this->verifone_auth_helper = new BasicAuth($this->verifone_settings);

        $this->client = new Client($this->api_url_mapper, $this->verifone_settings, $this->verifone_auth_helper, new SimpleCurl());
        $this->add_actions();
    }

    /**
     * Adds various actions to be used later on
     * Mostly hooks
     */
    public function add_actions() {
        add_action('woocommerce_update_options_payment_gateways_' . $this->id, [$this, 'process_admin_options',]);
        add_action('woocommerce_api_woo-verifone-key', [$this, 'process_key_request']);
        add_action('woocommerce_api_woo-verifone-3ds-jwt', [$this, 'process_3dsjwt_request']);
        add_action('woocommerce_api_woo-verifone-3ds-lookup', [$this, 'process_3dslookup_request']);
        add_action('admin_enqueue_scripts', [$this, 'admin_scripts']);
        add_action('wp_enqueue_scripts',[$this,'wp_enqueue_scripts']);
        add_action('wp_footer', [$this, 'wp_footer']);
        add_action('woocommerce_get_sections_checkout', [$this, 'add_subsection_inline']);
        add_action('woocommerce_order_item_add_action_buttons', [$this, 'verifone_checkout_capture'], 10, 2);
        add_action('woocommerce_api_woo-verifone-js', [ $this, 'process_js_checkout' ] );
        add_action('woocommerce_api_woo-verifone-inline-webhook', [$this, 'process_webhook']);
        add_action('woocommerce_api_manual_capture_payment', [$this, 'manual_capture_payment']);
        add_action( 'woocommerce_order_status_cancelled', [ $this, 'cancel_payment' ] );
        add_action( 'woocommerce_order_status_completed', [ $this, 'complete_payment' ] );
        add_action( 'woocommerce_order_status_refunded', [ $this, 'cancel_payment' ] );
    }

    /**
     * Includes a javascript file to be used in the backend
     */
    public function admin_scripts() {
        wp_enqueue_script('verifone_admin_script', plugins_url('assets/js/verifone_admin.js', dirname(__FILE__)));
    }

    public function init_form_fields() {
        require_once plugin_dir_path(__FILE__) . '../templates/init_form_fields.php';
        $current_subsection = (isset($_GET['vhc-tab']) && !empty($_GET['vhc-tab'])) ? $_GET['vhc-tab'] : '';
        switch ($current_subsection) {
            case 'verifone-account-settings-tab':
            default:
                $this->form_fields = getVerifoneInlineVerifoneAccountSettingsFields();
                break;
            case 'card-payment-settings-tab':
                $this->form_fields = getVerifoneInlineCardPaymentSettingsFields();
                break;
            case 'display-settings-tab':
                $this->form_fields = getVerifoneInlineDisplaySettingsFields();
                break;
            case 'advanced-settings-tab':
                $this->form_fields = getVerifoneInlineAdvancedSettingsFields();
                break;
        }
    }

    public function add_subsection_inline($settings_tabs) {
        $current_page = (isset($_GET['tab']) && !empty($_GET['tab'])) ? $_GET['tab'] : '';
        if ($current_page == 'checkout' && !self::$SUBSECTIONS_RENDERED) {
            $current_section = (isset($_GET['section']) && !empty($_GET['section'])) ? $_GET['section'] : '';
            if ($current_section == self::$PLUGIN_ID) {
                $sections = array(
                    'verifone-account-settings-tab' => __('Verifone account settings', 'woocommerce-verifone-account-settings-tab'),
                    'display-settings-tab' => __('Display settings', 'woocommerce-display-settings-tab'),
                    'card-payment-settings-tab' => __('Card payment settings', 'woocommerce-card-payment-settings-tab'),
                    'advanced-settings-tab' => __('Advanced settings', 'woocommerce-advanced-settings-tab'),
                );
                echo '<h2 style="display: inline-block">' . $this->method_title . '</h2>';
                wc_back_link(__('Return to payments', 'woocommerce'), admin_url('admin.php?page=wc-settings&tab=checkout'));
                echo '<p>' . $this->method_long_description . ' <a href="' . $this->method_description_link_url . '" target="_blank">' . $this->method_description_link_text . '</a></p>';
                echo '<p>' . __('Register this link to your Verifone Account', 'verifone-hosted') . '</p>
                    <div style="display: inline-block; padding: 5px 20px; border: 1px solid #BBBBBB; margin-bottom: 30px; border-radius: 2px; background-color: #EEEEEE;">
                    
                        <button style="margin-right: 10px; cursor: pointer; display: inline-block; border: 0" id="webhook_copy_button" type="button">
                            <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-clipboard" viewBox="0 0 16 16">
                              <path d="M4 1.5H3a2 2 0 0 0-2 2V14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V3.5a2 2 0 0 0-2-2h-1v1h1a1 1 0 0 1 1 1V14a1 1 0 0 1-1 1H3a1 1 0 0 1-1-1V3.5a1 1 0 0 1 1-1h1v-1z"/>
                              <path d="M9.5 1a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5h-3a.5.5 0 0 1-.5-.5v-1a.5.5 0 0 1 .5-.5h3zm-3-1A1.5 1.5 0 0 0 5 1.5v1A1.5 1.5 0 0 0 6.5 4h3A1.5 1.5 0 0 0 11 2.5v-1A1.5 1.5 0 0 0 9.5 0h-3z"/>
                            </svg>
                        </button>
                    <p style="display: inline-block" id="webhook_response_url">' . $this->get_webhook_response_url() . '</p>
                    </div>';
                echo '<br/>';
                echo '<ul style="margin-bottom: 20px" class="subsubsub">';
                $array_keys = array_keys($sections);
                $key = "vhc-tab";
                $current_subsection = (isset($_GET['vhc-tab']) && !empty($_GET['vhc-tab'])) ? $_GET['vhc-tab'] : 'default';
                foreach ($sections as $id => $label) {
                    $url = admin_url('admin.php?page=wc-settings&tab=checkout&section=verifone_inline_cart&' . $key . '=' . $id);
                    echo '<li><a href="' . esc_url($url) . '" class="' . ($current_subsection === $id || ($current_subsection === 'default' && $id === 'verifone-account-settings-tab') ? 'current' : '') . '">' . esc_html($label) . '</a> ' . (end($array_keys) === $id ? '' : '|') . ' </li>';
                }
                echo '</ul><br class="clear" />';
            }

            self::$SUBSECTIONS_RENDERED = true;
        }
    }

    public function wp_enqueue_scripts() {
        wp_enqueue_script('verifone_inline_online_script', 'https://cst.jsclient.vficloud.net/verifone.js',['jquery']);
        wp_enqueue_script('verifone_inline_script', plugins_url('assets/js/verifone_hosted_cart_inline.js', dirname(__FILE__)),['jquery']);
        wp_enqueue_style('verifone_inline_style', plugins_url('assets/css/verifone_hosted_cart_inline.css', dirname(__FILE__)));

        if ($this->trueOrFalse('_3ds_enable')) {
            if ($this->trueOrFalse('test'))
                wp_enqueue_script('verifone_inline_cardinal_script', 'https://songbirdstag.cardinalcommerce.com/edge/v1/songbird.js',['jquery','verifone_inline_script']);
            else
                wp_enqueue_script('verifone_inline_cardinal_script', 'https://songbird.cardinalcommerce.com/edge/v1/songbird.js',['jquery','verifone_inline_script']);
        }

        ?>
        <script type="text/javascript">
            const verifoneKeyUrl = "<?=WC()->api_request_url('woo-verifone-key')?>";
            const verifone3DSEnabled = <?=$this->trueOrFalse('_3ds_enable') ? 'true' : 'false'?>;
            const verifone3DSJwtUrl = "<?=WC()->api_request_url('woo-verifone-3ds-jwt')?>";
            const verifone3DSLookup = "<?=WC()->api_request_url('woo-verifone-3ds-lookup')?>";
            const verifoneJsBlocksCheckoutUrl = "<?=WC()->api_request_url('woo-verifone-js')?>";
        </script>
        <?php
    }

    public function payment_fields() {
        if(!$this->isWoocommerceBlocksAvailable()){
            echo '<div id="verifone_error"></div>';
            echo '<p>' . $this->get_option('description', __('Payments powered by', 'verifone-hosted') . ' <a href="https://www.verifone.com/" target="_blank">Verifone</a>') . '</p>';

            $wcPaymentGatewayCC = new \WC_Payment_Gateway_CC();
            $wcPaymentGatewayCC->id = 'verifone';
            $wcPaymentGatewayCC->payment_fields();
        }
    }

    public function wp_footer() {
        if($this->isWoocommerceBlocksAvailable()) {
            echo '<div id="verifone_wrapper" style="display:none;">';
            echo '<div id="verifone_error" style="display:none; background: rgba(255,0,0,0.3);padding: 6px;color: black;border-radius: 9px;"></div>';
            echo '<p>' . $this->get_option('description', __('Payments powered by', 'verifone-hosted') . ' <a href="https://www.verifone.com/" target="_blank">Verifone</a>') . '</p>';

            $wcPaymentGatewayCC = new \WC_Payment_Gateway_CC();
            $wcPaymentGatewayCC->id = 'verifone';
            $wcPaymentGatewayCC->payment_fields();
            echo '</div>';
        }
    }

    function process_js_checkout(){
        if ($_SERVER['REQUEST_METHOD'] !== 'POST' ||  $_POST['radio-control-wc-payment-method-options'] !== self::$PLUGIN_ID)
            $this->api_response_error();

        $cart = WC()->cart;
        $customer = WC()->customer;
        $order = new WC_Order();

        foreach ($cart->get_cart() as $cart_item_key => $cart_item) {
            $product_id = $cart_item['product_id'];
            $variation_id = $cart_item['variation_id'];
            $quantity = $cart_item['quantity'];

            // Add line items to the order
            $order->add_product(wc_get_product($product_id), $quantity, array(
                'variation_id' => $variation_id
            ));
        }
        
        $coupons=$cart->get_coupons();
        if($coupons){
            foreach ($coupons as $coupon){
                $order->apply_coupon($coupon);
            }
        }

        $order->set_address($customer->get_billing(), 'billing');
        $order->set_address($customer->get_shipping(), 'shipping');

        $order->calculate_totals();
        $order->save();
        
        return $this->api_response_ok($this->process_payment($order->get_id()));
    }

    public function process_payment($order_id) {
        $order = wc_get_order($order_id);

        if (!$order instanceof WC_Order) {
            $this->api_response_error('Invalid order');
        }

        $payment_gateways = WC()->payment_gateways->payment_gateways();
        if (!isset($payment_gateways[self::$PLUGIN_ID])) {
            $errorMsg = sprintf('Attempted to set payment method as %s however said payment method does not exist. This error should not happen unless the payment method is misspelled.',
                self::$PLUGIN_ID);
            $this->logger->debug($errorMsg);
            $this->api_response_error(__('There has been an error processing your order', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
        }

        $order->set_payment_method($payment_gateways[self::$PLUGIN_ID]);
        $order->save();

        global $woocommerce;

        $locale = $this->s40(get_locale(), 2);

        if (in_array($locale, self::NORWAY_CODES)) {
            $locale = "no";
        }

        //create the customer - will return NULL if the customer could not be created
        $customer = $this->getCustomer($order);
        if (!$customer) {
            $this->api_response_error(__('There has been an error processing your order', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
        }

        $transaction_mode = wp_is_mobile() ? 'P' : 'S';
        $orderAmount = $this->calculateAmount($order->get_total());

        $order->update_meta_data(VERIFONE_CAPTURE_MANUAL, $this->verifone_settings->isManualCapture() ? 'yes' : 'no');
        $order->update_meta_data(VERIFONE_ORDER_TRANSACTION_TYPE, $this->verifone_settings->getAuthType());
        $order->save_meta_data();

        $cardPayload = [
            'payment_provider_contract' => $this->verifone_settings->getPaymentContractId(),
            'amount' => $orderAmount,
            'customer' => $customer,
            'merchant_reference' => (string)$order->get_id(),
            'currency_code' => get_option('woocommerce_currency'),
            'encrypted_card' => $_POST['encrypted'],
            'public_key_alias' => $_POST['key_alias'],
            'auth_type' => $this->verifone_settings->getAuthType(),
            'capture_now' => !$this->verifone_settings->isManualCapture(),
        ];

        if ($this->trueOrFalse('_3ds_enable') && !empty($_POST['three_ds_lookup_data'])) {
            $threeDsLookupData = json_decode(strtr($_POST['three_ds_lookup_data'], ['\\"' => '"']), true);

            if (JSON_ERROR_NONE !== json_last_error()) {
                $this->api_response_error(__('There has been an error processing your order', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
            }

            $cardPayload['threed_authentication'] = [
                'eci_flag' => $threeDsLookupData['eci_flag'] ?? '',
                'enrolled' => $threeDsLookupData['enrolled'] ?? '',
                'pares_status' => $threeDsLookupData['pares_status'] ?? '',
                'threeds_version' => $threeDsLookupData['threeds_version'] ?? '',
                'ds_transaction_id' => $threeDsLookupData['ds_transaction_id'] ?? '',
                'signature_verification' => $threeDsLookupData['signature_verification'] ?? '',
            ];

            if (!empty($_POST['cca_jwt'])) {
                $threeDsJwtValidatepayload = [
                    'threeds_contract_id' => $this->get_option('_3ds_contract_id'),
                    'authentication_id' => $threeDsLookupData['authentication_id'],
                    'jwt' => $_POST['cca_jwt']
                ];

                try {
                    $response = $this->client->threeDSJwtValidate($threeDsJwtValidatepayload);
                } catch (\Exception $ex) {
                    $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
                    $this->api_response_error(__('There has been an error processing your order', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
                }

                if (0 != $response['error_no'])
                    $this->api_response_error(__('There has been an error processing your order', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);

                $enrolled = $response['validation_result']['enrolled'] ?? '';
                $paresStatus = $response['validation_result']['pares_status'] ?? '';
                $signatureVerification = $response['validation_result']['signature_verification'] ?? '';
                $acsUrl = $threeDsLookupData['acs_url'] ?? null;
                $payload = $threeDsLookupData['payload'] ?? null;

                $cardPayload['threed_authentication'] = [
                    'cavv' => $response['validation_result']['cavv'],
                    'error_desc' => $response['error_desc'],
                    'error_no' => $response['error_no'],
                ];
            } else {
                $enrolled = $threeDsLookupData['enrolled'] ?? '';
                $paresStatus = $threeDsLookupData['pares_status'] ?? '';
                $signatureVerification = $threeDsLookupData['signature_verification'] ?? '';
                $acsUrl = $threeDsLookupData['acs_url'] ?? null;
                $payload = $threeDsLookupData['payload'] ?? null;
            }

            //https://verifone.cloud/docs/uk-gateway/threeds_test_case_2_1_0
            if (
                ('Y' === $enrolled && 'N' === $paresStatus && 'Y' === $signatureVerification && empty($acsUrl) && empty($payload)) ||
                ('Y' === $enrolled && 'R' === $paresStatus && 'Y' === $signatureVerification && empty($acsUrl) && empty($payload)) ||
                ('Y' === $enrolled && 'C' === $paresStatus && 'Y' === $signatureVerification && !empty($acsUrl) && !empty($payload))
            ) {
                $this->api_response_error(__('Payment cannot be made with this card.', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
            }
        }

        try {
            $response = $this->client->postCard($cardPayload);
            if('FAILED'===$response['status'])
                throw new \Exception('Transaction failed');

            $transaction = $this->client->getTransaction($response['id']);
        } catch (\Exception $exception) {
            $this->logger->debug(
                'There was an error in the API response: ' . PHP_EOL .
                'Sent request: ' . json_encode($cardPayload) . PHP_EOL .
                'Received response: ' . $exception->getMessage() . PHP_EOL .
                'To endpoint: ' . $this->api_url_mapper->postCardUrl()
            );

            $this->api_response_error(__('Payment cannot be made with this card.', 'woocommerce-gateway-verifone-hosted-checkout'), __LINE__);
        }

        $order->update_meta_data(VERIFONE_PAYMENT_METHOD_USED, $response['payment_product']);
        $order->update_meta_data(VERIFONE_ORDER_TRANSACTION_ID, $response['id']);
        $order->add_order_note(__('Transaction ID: ' . $response['id']));

        if (in_array($transaction['transaction_status'], VERIFONE_TRANSACTION_STATUS_OK)) {
            if (VERIFONE_TRANSACTION_TYPE_SALE === $transaction['transaction_type']) {
                $order->payment_complete();

                if ($this->verifone_settings->getCompleteOrderOnPayment() || $this->isOrderOnlyVirtual($order)) {
                    //Cannot update status to completed for already active subscriptions.
                    if (!($order instanceof WC_Subscription) || 'active' !== $order->get_status()) {
                        $order->update_status('completed');
                    }
                }

                $order->add_order_note(__("Order payment is completed."), false);
            } else {
                if (!in_array($order->get_status(), ['completed', 'processing']) && !$this->noteExists($order->get_id(), '(transaction is in PRE/FINAL AUTH status)')) {
                    $order->add_order_note(__('The payment is not settled yet (transaction is in PRE/FINAL AUTH status), use the CAPTURE button to fully/partial capture the amount. 
                        Attempting to make a full refund, before capturing, will void the authorization and cancel the payment.'), false);
                    $order->update_status('processing');
                }
            }
        }

        $woocommerce->cart->empty_cart();

        $json_response = [
            'result' => 'success',
            'messages' => 'Order payment success',
            'refresh' => true,
            'reload' => false,
            'redirect' => $this->get_return_url($order)
        ];

        return $json_response;
    }

    public function process_refund($order_id, $amount = null, $reason = '') {
        $order = wc_get_order($order_id);

        if (VERIFONE_PAYMENT_METHOD_INLINE !== $order->get_payment_method())
            return;

        return $this->_process_refund($order_id, $amount, $reason);
    }

    /**
     * @param $order_id
     * @param string $type
     * @param null $amount
     *
     * @return void
     */
    protected function voidOrCapture( $order_id, string $type = 'capture', $amount = null ): void {
        // we make this check if the order is updated by changing the status and if 'manageOrders' is set to TRUE,
        // if we press the 'capture amount' button we don't make the check
        $calledBy = debug_backtrace()[1]['function'];
        if ( in_array( $calledBy, [ 'complete_payment', 'cancel_payment' ] ) && ! $this->verifone_settings->getManageOrders() ) {
            return;
        }
        $order          = wc_get_order( $order_id );
        $payment_method = $order->get_meta( VERIFONE_PAYMENT_METHOD_USED );

        if ( VERIFONE_PAYMENT_METHOD_INLINE===$order->get_payment_method() ) {
            if ( ! $order->get_meta( VERIFONE_CAPTURE_MANUAL ) ) {
                return;
            }

            if ( $order->get_meta( VERIFONE_CAPTURE_MANUAL ) && $order->get_meta( VERIFONE_CAPTURE_MANUAL ) !== 'yes' ) {
                return;
            }
            if ( $order->get_meta( VERIFONE_CAPTURE_DONE ) && $order->get_meta( VERIFONE_CAPTURE_DONE ) === 'yes' ) {
                return;
            }

            try {
                $transaction_id = $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID );
                if ( ! $amount ) {
                    $amount = $this->calculateAmount($order->get_total() );
                }
                $payload = [ 'amount' => $amount ];

                if ( $type == 'void' ) {
                    $result = $this->client->postVoidAuthorization( $payload, $transaction_id );
                }
                if ( $type == 'capture' ) {
                    if ( VERIFONE_PAYMENT_METHOD_PAYPAL === $payment_method ) {
                        $payload['amount'] = [
                            'value'        => $payload['amount'],
                            'currencyCode' => $order->get_currency(),
                        ];
                        $result            = $this->client->postPaypalCapture( $payload, $transaction_id );
                    } else {
                        $result = $this->client->postCapture( $payload, $transaction_id );
                    }
                }

                // a 400 request means the capture was already done
                // thus we can ignore it
                if ( isset( $result['code'] ) && $result['code'] == 400 ) {
                    return;
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_AUTHORIZED ) {
                    $message = sprintf( 'Order with transaction ID %s was authorized. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_SETTLEMENT_COMPLETED ) {
                    $message = sprintf( 'Order with transaction ID %s was settled. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_VOIDED ) {
                    $message = sprintf( 'Order with transaction ID %s was voided. Void ID %s', $transaction_id, $result['id'] );
                }

                if (isset($result['status']) && in_array($result['status'], VERIFONE_ENV_CHECKOUT_EVENT_FAILED)) {
                    $message = __("Payment was declined");
                    $order->update_status('failed');
                }

                if ( isset( $message ) ) {
                    $order->add_order_note( $message, false );
                    $order->update_meta_data( '__verifone_order_capture_done', 'yes' );
                } else {
                    $order->add_order_note( 'No change detected for order, please try setting the status again', false );
                }
                if ( is_callable( [ $order, 'save' ] ) ) {
                    $order->save();
                }
            } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
                // update the status to failed
                $order->update_status( 'failed', 'Unable to change order status, please try again later' );
                $this->logger->debug( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s . The payload: %s',
                    json_last_error(), __FILE__, __LINE__, $exception->getMessage(), json_encode($payload) ) );
            }
        }
    }

    public function process_key_request() {
        try {
            $response = $this->client->getKeys();
        } catch (\Exception $ex) {
            $response = json_decode($ex->getMessage(), true);
            if (json_last_error() !== JSON_ERROR_NONE || VERIFONE_KEYS_NOT_FOUND !== $response['details']['error']) {
                $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
                $this->api_response_error('Unable to process data', __LINE__);
            }

            try {
                $response = $this->client->generateKey();
            } catch (\Exception $ex) {
                $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
                $this->api_response_error('Unable to process data', __LINE__);
            }

            $this->api_response_ok($response);
        }

        if (empty($response['public_keys'])) {
            $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
            $this->api_response_error('Unable to process data', __LINE__);
        }

        $keys = $response['public_keys'];
        shuffle($keys);
        $this->api_response_ok($keys[0]);
    }

    public function process_3dsjwt_request() {
        if ($_SERVER['REQUEST_METHOD'] !== 'POST')
            $this->api_response_error();

        try {
            $response = $this->client->generateJwt($this->get_option('_3ds_contract_id'));
        } catch (\Exception $ex) {
            $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
            $this->api_response_error('Unable to process data', __LINE__);
        }

        $this->api_response_ok($response['jwt']);
    }

    public function process_3dslookup_request() {
        if ($_SERVER['REQUEST_METHOD'] !== 'POST')
            $this->api_response_error();

        $input = file_get_contents('php://input');
        parse_str($input, $inputData);

        $customer = WC()->customer;

        $payload = [
            'threeds_contract_id' => $this->get_option('_3ds_contract_id'),
            'amount' => $this->calculateAmount($inputData['amount']),
            'currency_code' => get_option('woocommerce_currency'),
            'billing_first_name' => $inputData['billing_first_name'] ?? $customer->get_billing_first_name(),
            'billing_last_name' => $inputData['billing_last_name'] ?? $customer->get_billing_last_name(),
            'billing_address_1' => $inputData['billing_address_1'] ?? $customer->get_billing_address_1(),
            'billing_city' => $inputData['billing_city'] ?? $customer->get_billing_city(),
            'billing_country_code' => $inputData['billing_country'] ?? $customer->get_billing_country(),
            'email' => $inputData['email'] ?? $customer->get_billing_email(),
            'device_info_id' => $inputData['sessionId'],
            'encrypted_card' => $inputData['encrypted'],
            'public_key_alias' => $inputData['key_alias'],
        ];

        try {
            $response = $this->client->threeDSLookup($payload);
        } catch (\Exception $ex) {
            $this->logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API: ' . $ex->getMessage());
            $this->api_response_error('Unable to process data: ' . $ex->getMessage(), __LINE__);
        }

        $this->api_response_ok($response);
    }

    function getPluginId(): string {
        return self::$PLUGIN_ID;
    }

    function get_webhook_response_url() {
        return WC()->api_request_url('woo-verifone-inline-webhook');
    }
}
