<?php
use Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType;
use Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry;

defined( 'ABSPATH' ) || exit;
require_once 'class-verifone-hosted-constants.php';
/**
 *
 *
 * @extends AbstractPaymentMethodType
 */
final class WC_Verifone_Hosted_Blocks_Support extends AbstractPaymentMethodType
{
    /**
     * @var WC_Logger
     */
    protected $logger;

    /**
     * Payment method name defined by payment methods extending this class.
     *
     * @var string
     */
    protected $name = 'verifone_hosted';

    public function __construct() {
        $this->logger = new WC_Logger();
    }

    public function initialize()
    {
        add_action( 'woocommerce_blocks_loaded', 'verifone_hosted_woocommerce_blocks_support' );
    }

    public function is_active() {
        return true;
    }

    function verifone_hosted_woocommerce_blocks_support() {
        if ( class_exists( 'Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType' ) ) {
            add_action(
                'woocommerce_blocks_payment_method_type_registration',
                function( PaymentMethodRegistry $payment_method_registry ) {
                    $payment_method_registry->register( new WC_Verifone_Hosted_Blocks_Support() );
                }
            );
        }
    }

    /**
     * Returns an array of scripts/handles to be registered for this payment method.
     *
     * @return array
     */
    public function get_payment_method_script_handles() {
        $asset_path   =dirname(__FILE__). '/../assets/js/blocks/frontend/blocks.asset.php';
        $version      = WC_VERIFONE_VERSION;
        $dependencies = [];
        if ( file_exists( $asset_path ) ) {
            $asset        = require $asset_path;
            $version      = is_array( $asset ) && isset( $asset['version'] )
                ? $asset['version']
                : $version;
            $dependencies = is_array( $asset ) && isset( $asset['dependencies'] )
                ? $asset['dependencies']
                : $dependencies;
        }

        wp_register_script(
            'wc-verifone-blocks-integration',
            plugins_url('assets/js/blocks/frontend/blocks.js', dirname(__FILE__)),
            $dependencies,
            $version,
            true
        );

        return [ 'wc-verifone-blocks-integration' ];
    }

    /**
     * Returns an array of key=>value pairs of data made available to the payment methods script.
     *
     * @return array
     */
    public function get_payment_method_data() {
        $payment_gateways_class = WC()->payment_gateways();
        $payment_gateways = $payment_gateways_class->payment_gateways();

        $verifoneHostedCartGateway = $payment_gateways['verifone_hosted_cart'];
        $verifoneInlineCartGateway = $payment_gateways['verifone_inline_cart'];

        return [
            'hosted' => [
                'enabled' => 'yes' === $verifoneHostedCartGateway->get_option('enabled') ? true : false,
                'title' => $verifoneHostedCartGateway->get_option('title'),
                'description' => $verifoneHostedCartGateway->get_option('description'),
                'checkout_button_text' => $verifoneHostedCartGateway->get_option('checkout_button_text')?:$verifoneHostedCartGateway->get_option('title'),
            ],
            'inline' => [
                'enabled' => 'yes' === $verifoneInlineCartGateway->get_option('enabled') ? true : false,
                'title' => $verifoneInlineCartGateway->get_option('title'),
                'description' => $verifoneInlineCartGateway->get_option('description'),
                'checkout_button_text' => $verifoneInlineCartGateway->get_option('checkout_button_text')?:$verifoneInlineCartGateway->get_option('title'),
            ]
        ];
    }

}