<?php
/**
 * Plugin Name: Verifone Hosted Cart
 * Plugin URI:
 * Description: Allows you to use Verifone Hosted payment gateway with the WooCommerce plugin.
 * Version: 3.1.2
 * Author: verifone
 * Author URI: https://www.verifone.cloud
 * WC requires at least: 4.6.2
 * WC tested up to: 8.4.0
 * Requires PHP: 7.4
 */

if ( ! defined( 'ABSPATH' ) ) {
    exit; // Exit if accessed directly
}
add_filter( 'woocommerce_payment_gateways', 'add_verifone_hosted_gateway' );
function add_verifone_hosted_gateway( $methods ) {
    require_once 'autoload.php';
    require_once 'bootstrap.php';
    $methods[] = 'WC_Verifone_Hosted_Gateway';
    $methods[] = 'WC_Verifone_Inline_Gateway';

    return $methods;
}

//for invoice4u we need to break the line items into 1 item per line item
if ( !empty(get_option( 'woocommerce_verifone_hosted_cart_settings' )['invoice4u_enabled'])
     && get_option( 'woocommerce_verifone_hosted_cart_settings' )['invoice4u_enabled'] === 'yes'
     && get_option( 'woocommerce_verifone_hosted_cart_settings' )['card_enabled'] === 'yes') {

    add_filter( 'woocommerce_add_cart_item_data', 'verifone_split_product_individual_cart_items', 10, 2 );
    add_filter( 'woocommerce_is_sold_individually', '__return_true' );
}
function verifone_split_product_individual_cart_items( $cart_item_data, $product_id ) {

    $cart_item_data['unique_key'] = uniqid();

    return $cart_item_data;
}

add_action( 'before_woocommerce_init', function() {
    if ( class_exists( \Automattic\WooCommerce\Utilities\FeaturesUtil::class ) ) {
        \Automattic\WooCommerce\Utilities\FeaturesUtil::declare_compatibility( 'custom_order_tables', 'verifone-hosted/wc-verifone-hosted.php', true );
    }
} );

add_action( 'woocommerce_blocks_loaded', 'woocommerce_gateway_verifone_woocommerce_block_support' );

function woocommerce_gateway_verifone_woocommerce_block_support() {
    if ( class_exists( 'Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType' ) ) {
        require_once dirname( __FILE__ ) . '/includes/class-wc-verifone-hosted-blocks-support.php';

        add_action(
            'woocommerce_blocks_payment_method_type_registration',
            function( Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry $payment_method_registry ) {
                $container = Automattic\WooCommerce\Blocks\Package::container();
                $container->register(
                    WC_Verifone_Hosted_Blocks_Support::class,
                    function() {
                        return new WC_Verifone_Hosted_Blocks_Support();
                    }
                );
                $payment_method_registry->register(
                    $container->get( WC_Verifone_Hosted_Blocks_Support::class )
                );
            },
            5
        );
    }
}

